@@ -188,7 +188,7 @@ main(int argc, char *argv[])
 		(void) printf(_("%s grammar version %d\n"), getprogname(), SUDOERS_GRAMMAR_VERSION);
 		goto done;
 	    case 'c':
-		checkonly++;		/* check mode */
+		checkonly = true;	/* check mode */
 		break;
 	    case 'f':
 		sudoers_path = optarg;	/* sudoers file path */
@@ -198,10 +198,10 @@ main(int argc, char *argv[])
 		help();
 		break;
 	    case 's':
-		strict++;		/* strict mode */
+		strict = true;		/* strict mode */
 		break;
 	    case 'q':
-		quiet++;		/* quiet mode */
+		quiet = false;		/* quiet mode */
 		break;
 	    default:
 		usage(1);
@@ -235,7 +235,7 @@ main(int argc, char *argv[])
     if ((yyin = open_sudoers(sudoers_path, true, NULL)) == NULL) {
 	error(1, "%s", sudoers_path);
     }
-    init_parser(sudoers_path, 0);
+    init_parser(sudoers_path, false);
     yyparse();
     (void) update_defaults(SETDEF_GENERIC|SETDEF_HOST|SETDEF_USER);
 
@@ -472,6 +472,9 @@ reparse_sudoers(char *editor, char *args, bool strict, bool quiet)
     int ch;
     debug_decl(reparse_sudoers, SUDO_DEBUG_UTIL)
 
+    if (tq_empty(&sudoerslist))
+	debug_return;
+
     /*
      * Parse the edited sudoers files and do sanity checking
      */
@@ -497,10 +500,10 @@ reparse_sudoers(char *editor, char *args, bool strict, bool quiet)
 	}
 	fclose(yyin);
 	if (!parse_error) {
-	    if (!update_defaults(SETDEF_GENERIC|SETDEF_HOST|SETDEF_USER) ||
+	    if (!check_defaults(SETDEF_ALL, quiet) ||
 		check_aliases(strict, quiet) != 0) {
 		parse_error = true;
-		errorfile = sp->path;
+		errorfile = NULL;
 	    }
 	}
 
@@ -524,10 +527,11 @@ reparse_sudoers(char *editor, char *args, bool strict, bool quiet)
 	    tq_foreach_fwd(&sudoerslist, sp) {
 		if (errorfile == NULL || strcmp(sp->path, errorfile) == 0) {
 		    edit_sudoers(sp, editor, args, errorlineno);
-		    break;
+		    if (errorfile != NULL)
+			break;
 		}
 	    }
-	    if (sp == NULL) {
+	    if (errorfile != NULL && sp == NULL) {
 		errorx(1, _("internal error, unable to find %s in list!"),
 		    sudoers);
 	    }
@@ -822,9 +826,12 @@ check_syntax(char *sudoers_path, bool quiet, bool strict, bool oldperms)
 	parse_error = true;
 	errorfile = sudoers_path;
     }
-    if (!parse_error && check_aliases(strict, quiet) != 0) {
-	parse_error = true;
-	errorfile = sudoers_path;
+    if (!parse_error) {
+	if (!check_defaults(SETDEF_ALL, quiet) ||
+	    check_aliases(strict, quiet) != 0) {
+	    parse_error = true;
+	    errorfile = NULL;
+	}
     }
     ok = !parse_error;
 
@@ -833,7 +840,7 @@ check_syntax(char *sudoers_path, bool quiet, bool strict, bool oldperms)
 	    if (errorlineno != -1)
 		(void) printf(_("parse error in %s near line %d\n"),
 		    errorfile, errorlineno);
-	    else
+	    else if (errorfile != NULL)
 		(void) printf(_("parse error in %s\n"), errorfile);
 	}
     } else {
@@ -1068,18 +1075,16 @@ alias_remove_recursive(char *name, int type)
     bool rval = true;
     debug_decl(alias_remove_recursive, SUDO_DEBUG_ALIAS)
 
-    if ((a = alias_find(name, type)) != NULL) {
+    if ((a = alias_remove(name, type)) != NULL) {
 	tq_foreach_fwd(&a->members, m) {
 	    if (m->type == ALIAS) {
 		if (!alias_remove_recursive(m->name, type))
 		    rval = false;
 	    }
 	}
+	rbinsert(alias_freelist, a);
     }
     alias_seqno++;
-    a = alias_remove(name, type);
-    if (a)
-	rbinsert(alias_freelist, a);
     debug_return_bool(rval);
 }
 
